"""
\3 -> every third for NUCLEOTIDE
tests/data/chars/primates.chars.subsets-all.nexus
begin sets;
>-------charset coding = 2-457 660-896;
>-------charset noncoding = 1 458-659 897-898;
>-------charset 1stpos = 2-457\3 660-896\3;
>-------charset 2ndpos = 3-457\3 661-896\3;
>-------charset 3rdpos = 4-457\3 662-.\3;
end;

tests/data/chars/orti.nex
BEGIN SETS;
   TaxSet EPAC = 1-14;
   TaxSet WPAC = 15-23;
END;

tests/data/chars/COII_Apes.nex
BEGIN SETS;
   TaxSet Human = 1-6;
   TaxSet Chimp = 7-11;
   TaxSet PygmyChimp = 12-15;
   TaxSet Gorilla = 16-21;
   TaxSet Orangutan = 22-24;
END;
"""